use std::{future::Future, marker::PhantomData};

use super::{Device2 as Device, Queue};
use ash::{prelude::*, vk};

pub struct FenceFuture<'a> {
    device: Device,
    fence: vk::Fence,
    _pd: PhantomData<&'a ()>,
}

impl Drop for FenceFuture<'_> {
    fn drop(&mut self) {
        unsafe { self.device.dev().destroy_fence(self.fence, None) };
    }
}

impl FenceFuture<'_> {
    /// Unsafe because `fence` must not be destroyed while this future is live.
    pub unsafe fn from_fence(device: Device, fence: vk::Fence) -> Self {
        Self {
            device,
            fence,
            _pd: PhantomData,
        }
    }
    pub fn block_until(&self) -> VkResult<()> {
        unsafe {
            self.device
                .dev()
                .wait_for_fences(&[self.fence], true, u64::MAX)
        }
    }
}

impl Future for FenceFuture<'_> {
    type Output = ();

    fn poll(
        self: std::pin::Pin<&mut Self>,
        cx: &mut std::task::Context<'_>,
    ) -> std::task::Poll<Self::Output> {
        let signaled = unsafe {
            self.device
                .dev()
                .get_fence_status(self.fence)
                .unwrap_or(false)
        };

        if signaled {
            std::task::Poll::Ready(())
        } else {
            tokio::task::spawn_blocking(move || {});
            std::task::Poll::Pending
        }
    }
}

pub struct SingleUseCommandPool {
    device: Device,
    pool: vk::CommandPool,
}

impl Drop for SingleUseCommandPool {
    fn drop(&mut self) {
        unsafe {
            self.device.dev().destroy_command_pool(self.pool, None);
        }
    }
}

impl SingleUseCommandPool {
    pub fn new(device: Device, family_index: u32) -> VkResult<Self> {
        let pool_info = vk::CommandPoolCreateInfo::default()
            .queue_family_index(family_index)
            .flags(vk::CommandPoolCreateFlags::TRANSIENT);

        let pool =
            unsafe { device.dev().create_command_pool(&pool_info, None)? };

        Ok(Self { device, pool })
    }

    pub fn pool(&self) -> vk::CommandPool {
        self.pool
    }
}

pub struct SingleUseCommand {
    device: Device,
    pool: vk::CommandPool,
    buffer: vk::CommandBuffer,
}

impl Drop for SingleUseCommand {
    fn drop(&mut self) {
        unsafe {
            self.device
                .dev()
                .free_command_buffers(self.pool, &[self.buffer])
        };
    }
}

impl SingleUseCommand {
    pub fn new(device: Device, pool: vk::CommandPool) -> VkResult<Self> {
        let buffer = unsafe {
            let alloc_info = vk::CommandBufferAllocateInfo::default()
                .command_buffer_count(1)
                .command_pool(pool)
                .level(vk::CommandBufferLevel::PRIMARY);
            let buffer = device.dev().allocate_command_buffers(&alloc_info)?[0];

            let begin_info = vk::CommandBufferBeginInfo::default()
                .flags(vk::CommandBufferUsageFlags::ONE_TIME_SUBMIT);
            device.dev().begin_command_buffer(buffer, &begin_info)?;

            buffer
        };
        Ok(Self {
            device,
            pool,
            buffer,
        })
    }

    pub fn command_buffer(&self) -> vk::CommandBuffer {
        self.buffer
    }

    fn submit_fence(
        &self,
        queue: Queue,
        wait: Option<(vk::Semaphore, vk::PipelineStageFlags)>,
        signal: Option<vk::Semaphore>,
    ) -> VkResult<vk::Fence> {
        unsafe { self.device.dev().end_command_buffer(self.buffer)? };

        let buffers = [self.buffer];
        let mut submit_info =
            vk::SubmitInfo::default().command_buffers(&buffers);

        if let Some(semaphore) = signal.as_ref() {
            // SAFETY: T and [T;1] have the same layout
            submit_info = submit_info.signal_semaphores(unsafe {
                core::mem::transmute::<&vk::Semaphore, &[vk::Semaphore; 1]>(
                    semaphore,
                )
            });
        }
        if let Some((semaphore, stage)) = wait.as_ref() {
            submit_info = submit_info
                .wait_semaphores(core::slice::from_ref(semaphore))
                .wait_dst_stage_mask(core::slice::from_ref(stage));
        }

        let fence = {
            let create_info = vk::FenceCreateInfo::default();
            unsafe { self.device.dev().create_fence(&create_info, None)? }
        };

        queue.with_locked(|queue| unsafe {
            self.device.dev().queue_submit(queue, &[submit_info], fence)
        })?;

        Ok(fence)
    }

    pub fn submit_async<'a>(
        &'a self,
        queue: Queue,
        wait: Option<(vk::Semaphore, vk::PipelineStageFlags)>,
        signal: Option<vk::Semaphore>,
    ) -> VkResult<FenceFuture<'a>> {
        let device = self.device.clone();
        let fence = self.submit_fence(queue, wait, signal)?;

        Ok(unsafe { FenceFuture::from_fence(device, fence) })
    }

    pub fn submit_blocking(
        self,
        queue: Queue,
        wait: Option<(vk::Semaphore, vk::PipelineStageFlags)>,
        signal: Option<vk::Semaphore>,
    ) -> VkResult<()> {
        let device = self.device.clone();
        let fence = self.submit_fence(queue, wait, signal)?;

        unsafe {
            device.dev().wait_for_fences(&[fence], true, u64::MAX)?;
            device.dev().destroy_fence(fence, None);
        }
        Ok(())
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    async fn async_submit(cmd: SingleUseCommand, queue: Queue) {
        cmd.submit_async(queue, None, None).unwrap().await;
    }
}
